@@ -1,45 +1,45 @@
-//! The module implements [`HashMap`].
+//! [`HashMap`] is a concurrent and asynchronous hash map.
 
-use super::async_yield::{self, AwaitableBarrier};
 use super::ebr::{Arc, AtomicArc, Barrier};
 use super::hash_table::cell::{Locker, Reader};
 use super::hash_table::cell_array::CellArray;
 use super::hash_table::HashTable;
+use super::wait_queue::AsyncWait;
 
 use std::borrow::Borrow;
 use std::collections::hash_map::RandomState;
 use std::hash::{BuildHasher, Hash};
+use std::pin::Pin;
 use std::sync::atomic::Ordering::{Acquire, Relaxed};
 use std::sync::atomic::{AtomicU8, AtomicUsize};
 
 /// Scalable concurrent hash map.
 ///
-/// [`HashMap`] is a concurrent hash map data structure that is targeted at a highly concurrent
-/// workload. The use of an epoch-based reclamation technique enables the data structure to
-/// implement non-blocking resizing and fine-granular locking. A [`HashMap`] instance only has a
-/// single array of entries instead of a fixed number of lock-protected hash tables. An entry of
-/// the array is called a `Cell`; it manages a fixed number of key-value pairs using a customized
-/// mutex in it, and resolves hash conflicts by allocating a linked list of smaller hash tables.
+/// [`HashMap`] is a concurrent and asynchronous hash map data structure that is targeted at a
+/// highly concurrent workload. The use of an epoch-based reclamation technique enables the hash
+/// map to implement non-blocking resizing and fine-granular locking. A [`HashMap`] instance has a
+/// single array of *buckets* instead of a fixed number of lock-protected hash tables. Each bucket
+/// has a fixed size array of key-value pairs and a customized mutex to protect the data, and it
+/// resolves hash conflicts by allocating a linked list of bucket-local hash tables.
 ///
 /// ## The key features of [`HashMap`]
 ///
 /// * Non-sharded: the data is stored in a single array of key-value pairs.
 /// * Non-blocking resizing: resizing does not block other threads or tasks.
 /// * Automatic resizing: it automatically grows or shrinks.
-/// * Incremental resizing: each access to the data structure is mandated to rehash a fixed
-///   number of key-value pairs.
-/// * Optimized resizing: key-value pairs managed by a single `Cell` are guaranteed to be
-///   relocated to consecutive `Cell` instances.
-/// * No busy waiting: the customized mutex never spins.
-/// * Linearlizability: [`HashMap`] methods are linearlizable.
+/// * Incremental resizing: each access to the hash map is mandated to move a fixed
+///   number of key-value pairs if an old array is present.
+/// * No busy waiting: the thread or asynchronous task is suspended until the desired resource
+///   becomes available.
+/// * Linearizability: [`HashMap`] methods are linearizable.
 ///
 /// ## The key statistics for [`HashMap`]
 ///
 /// * The expected size of metadata for a single key-value pair: 2-byte.
 /// * The expected number of atomic write operations required for an operation on a single key: 2.
 /// * The expected number of atomic variables accessed during a single key operation: 2.
-/// * The number of entries managed by a single metadata `Cell` without a linked list: 32.
-/// * The expected maximum linked list length when resize is triggered: log(capacity) / 8.
+/// * The number of entries managed by a single bucket without a linked list: 32.
+/// * The expected maximum linked list length when a resize is triggered: log(capacity) / 8.
 pub struct HashMap<K, V, H = RandomState>
 where
     K: 'static + Eq + Hash + Sync,
@@ -155,11 +155,6 @@ where
     ///
     /// Returns an error along with the supplied key-value pair if the key exists.
     ///
-    /// # Panics
-    ///
-    /// Panics if memory allocation fails, or the number of entries in the target cell reaches
-    /// `u32::MAX`.
-    ///
     /// # Examples
     ///
     /// ```
@@ -173,10 +168,8 @@ where
     #[inline]
     pub fn insert(&self, key: K, val: V) -> Result<(), (K, V)> {
         let (hash, partial_hash) = self.hash(&key);
-        if let Some((k, v)) = self
-            .insert_entry::<false>(key, val, hash, partial_hash, &Barrier::new())
-            .ok()
-            .unwrap()
+        if let Ok(Some((k, v))) =
+            self.insert_entry(key, val, hash, partial_hash, None, &Barrier::new())
         {
             Err((k, v))
         } else {
@@ -186,7 +179,7 @@ where
 
     /// Inserts a key-value pair into the [`HashMap`].
     ///
-    /// It is an asynchronous method returning an `impl Future` for the caller to await or poll.
+    /// It is an asynchronous method returning an `impl Future` for the caller to await.
     ///
     /// # Errors
     ///
@@ -204,7 +197,16 @@ where
     pub async fn insert_async(&self, mut key: K, mut val: V) -> Result<(), (K, V)> {
         let (hash, partial_hash) = self.hash(&key);
         loop {
-            match self.insert_entry::<true>(key, val, hash, partial_hash, &Barrier::new()) {
+            let mut async_wait = AsyncWait::default();
+            let mut async_wait_pinned = Pin::new(&mut async_wait);
+            match self.insert_entry(
+                key,
+                val,
+                hash,
+                partial_hash,
+                Some(async_wait_pinned.mut_ptr()),
+                &Barrier::new(),
+            ) {
                 Ok(Some(returned)) => return Err(returned),
                 Ok(None) => return Ok(()),
                 Err(returned) => {
@@ -212,7 +214,7 @@ where
                     val = returned.1;
                 }
             }
-            async_yield::async_yield().await;
+            async_wait_pinned.await;
         }
     }
 
@@ -242,7 +244,7 @@ where
         let (hash, partial_hash) = self.hash(key_ref);
         let barrier = Barrier::new();
         let (_, _locker, iterator) = self
-            .acquire::<Q, false>(key_ref, hash, partial_hash, &barrier)
+            .acquire::<Q>(key_ref, hash, partial_hash, None, &barrier)
             .ok()?;
         if let Some(iterator) = iterator {
             if let Some((k, v)) = iterator.get() {
@@ -257,7 +259,7 @@ where
     /// Updates an existing key-value pair.
     ///
     /// It returns `None` if the key does not exist. It is an asynchronous method returning an
-    /// `impl Future` for the caller to await or poll.
+    /// `impl Future` for the caller to await.
     ///
     /// # Examples
     ///
@@ -278,9 +280,15 @@ where
     {
         let (hash, partial_hash) = self.hash(key_ref);
         loop {
-            if let Ok((_, _locker, iterator)) =
-                self.acquire::<Q, true>(key_ref, hash, partial_hash, &Barrier::new())
-            {
+            let mut async_wait = AsyncWait::default();
+            let mut async_wait_pinned = Pin::new(&mut async_wait);
+            if let Ok((_, _locker, iterator)) = self.acquire::<Q>(
+                key_ref,
+                hash,
+                partial_hash,
+                Some(async_wait_pinned.mut_ptr()),
+                &Barrier::new(),
+            ) {
                 if let Some(iterator) = iterator {
                     if let Some((k, v)) = iterator.get() {
                         #[allow(clippy::cast_ref_to_mut)]
@@ -289,17 +297,12 @@ where
                 }
                 return None;
             }
-            async_yield::async_yield().await;
+            async_wait_pinned.await;
         }
     }
 
     /// Constructs the value in-place, or modifies an existing value corresponding to the key.
     ///
-    /// # Panics
-    ///
-    /// Panics if memory allocation fails, or the number of entries in the target cell is
-    /// reached `u32::MAX`.
-    ///
     /// # Examples
     ///
     /// ```
@@ -321,29 +324,24 @@ where
     ) {
         let (hash, partial_hash) = self.hash(&key);
         let barrier = Barrier::new();
-        let (_, locker, iterator) = self
-            .acquire::<_, false>(&key, hash, partial_hash, &barrier)
-            .ok()
-            .unwrap();
-        if let Some(iterator) = iterator {
-            if let Some((k, v)) = iterator.get() {
-                // The presence of `locker` prevents the entry from being modified outside it.
-                #[allow(clippy::cast_ref_to_mut)]
-                updater(k, unsafe { &mut *(v as *const V as *mut V) });
-                return;
+        if let Ok((_, locker, iterator)) =
+            self.acquire::<_>(&key, hash, partial_hash, None, &barrier)
+        {
+            if let Some(iterator) = iterator {
+                if let Some((k, v)) = iterator.get() {
+                    // The presence of `locker` prevents the entry from being modified outside it.
+                    #[allow(clippy::cast_ref_to_mut)]
+                    updater(k, unsafe { &mut *(v as *const V as *mut V) });
+                    return;
+                }
             }
-        }
-        locker.insert(key, constructor(), partial_hash, &barrier);
+            locker.insert(key, constructor(), partial_hash, &barrier);
+        };
     }
 
     /// Constructs the value in-place, or modifies an existing value corresponding to the key.
     ///
-    /// It is an asynchronous method returning an `impl Future` for the caller to await or poll.
-    ///
-    /// # Panics
-    ///
-    /// Panics if memory allocation fails, or the number of entries in the target cell is
-    /// reached `u32::MAX`.
+    /// It is an asynchronous method returning an `impl Future` for the caller to await.
     ///
     /// # Examples
     ///
@@ -363,9 +361,15 @@ where
     ) {
         let (hash, partial_hash) = self.hash(&key);
         loop {
-            if let Ok((_, locker, iterator)) =
-                self.acquire::<_, true>(&key, hash, partial_hash, &Barrier::new())
-            {
+            let mut async_wait = AsyncWait::default();
+            let mut async_wait_pinned = Pin::new(&mut async_wait);
+            if let Ok((_, locker, iterator)) = self.acquire::<_>(
+                &key,
+                hash,
+                partial_hash,
+                Some(async_wait_pinned.mut_ptr()),
+                &Barrier::new(),
+            ) {
                 if let Some(iterator) = iterator {
                     if let Some((k, v)) = iterator.get() {
                         #[allow(clippy::cast_ref_to_mut)]
@@ -376,7 +380,7 @@ where
                 locker.insert(key, constructor(), partial_hash, &Barrier::new());
                 return;
             }
-            async_yield::async_yield().await;
+            async_wait_pinned.await;
         }
     }
 
@@ -404,7 +408,7 @@ where
 
     /// Removes a key-value pair if the key exists.
     ///
-    /// It is an asynchronous method returning an `impl Future` for the caller to await or poll.
+    /// It is an asynchronous method returning an `impl Future` for the caller to await.
     ///
     /// # Examples
     ///
@@ -448,11 +452,12 @@ where
         Q: Eq + Hash + ?Sized,
     {
         let (hash, partial_hash) = self.hash(key_ref);
-        self.remove_entry::<Q, _, false>(
+        self.remove_entry::<Q, _>(
             key_ref,
             hash,
             partial_hash,
             &mut condition,
+            None,
             &Barrier::new(),
         )
         .ok()
@@ -461,7 +466,7 @@ where
 
     /// Removes a key-value pair if the key exists and the given condition is met.
     ///
-    /// It is an asynchronous method returning an `impl Future` for the caller to await or poll.
+    /// It is an asynchronous method returning an `impl Future` for the caller to await.
     ///
     /// # Examples
     ///
@@ -484,16 +489,19 @@ where
     {
         let (hash, partial_hash) = self.hash(key_ref);
         loop {
-            if let Ok(result) = self.remove_entry::<Q, F, true>(
+            let mut async_wait = AsyncWait::default();
+            let mut async_wait_pinned = Pin::new(&mut async_wait);
+            if let Ok(result) = self.remove_entry::<Q, F>(
                 key_ref,
                 hash,
                 partial_hash,
                 &mut condition,
+                Some(async_wait_pinned.mut_ptr()),
                 &Barrier::new(),
             ) {
                 return result.0;
             }
-            async_yield::async_yield().await;
+            async_wait_pinned.await;
         }
     }
 
@@ -513,19 +521,22 @@ where
     /// assert_eq!(hashmap.read(&1, |_, v| *v).unwrap(), 10);
     /// ```
     #[inline]
-    pub fn read<Q, R, F: FnMut(&K, &V) -> R>(&self, key_ref: &Q, reader: F) -> Option<R>
+    pub fn read<Q, R, F: FnMut(&K, &V) -> R>(&self, key_ref: &Q, mut reader: F) -> Option<R>
     where
         K: Borrow<Q>,
         Q: Eq + Hash + ?Sized,
     {
         let barrier = Barrier::new();
-        self.read_with(key_ref, reader, &barrier)
+        let (hash, partial_hash) = self.hash(key_ref);
+        self.read_entry::<Q, R, F>(key_ref, hash, partial_hash, &mut reader, None, &barrier)
+            .ok()
+            .and_then(|r| r)
     }
 
     /// Reads a key-value pair.
     ///
     /// It returns `None` if the key does not exist. It is an asynchronous method returning an
-    /// `impl Future` for the caller to await or poll.
+    /// `impl Future` for the caller to await.
     ///
     /// # Examples
     ///
@@ -548,55 +559,22 @@ where
     {
         let (hash, partial_hash) = self.hash(key_ref);
         loop {
-            if let Ok(result) = self.read_entry::<Q, R, _, true>(
+            let mut async_wait = AsyncWait::default();
+            let mut async_wait_pinned = Pin::new(&mut async_wait);
+            if let Ok(result) = self.read_entry::<Q, R, _>(
                 key_ref,
                 hash,
                 partial_hash,
                 &mut reader,
+                Some(async_wait_pinned.mut_ptr()),
                 &Barrier::new(),
             ) {
                 return result;
             }
-            async_yield::async_yield().await;
+            async_wait_pinned.await;
         }
     }
 
-    /// Reads a key-value pair using the supplied [`Barrier`].
-    ///
-    /// It enables the caller to use the value reference outside the method. It returns `None`
-    /// if the key does not exist.
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// use scc::ebr::Barrier;
-    /// use scc::HashMap;
-    ///
-    /// let hashmap: HashMap<u64, u32> = HashMap::default();
-    ///
-    /// assert!(hashmap.insert(1, 10).is_ok());
-    ///
-    /// let barrier = Barrier::new();
-    /// let value_ref = hashmap.read_with(&1, |k, v| v, &barrier).unwrap();
-    /// assert_eq!(*value_ref, 10);
-    /// ```
-    #[inline]
-    pub fn read_with<'b, Q, R, F: FnMut(&'b K, &'b V) -> R>(
-        &self,
-        key_ref: &Q,
-        mut reader: F,
-        barrier: &'b Barrier,
-    ) -> Option<R>
-    where
-        K: Borrow<Q>,
-        Q: Eq + Hash + ?Sized,
-    {
-        let (hash, partial_hash) = self.hash(key_ref);
-        self.read_entry::<Q, R, F, false>(key_ref, hash, partial_hash, &mut reader, barrier)
-            .ok()
-            .and_then(|r| r)
-    }
-
     /// Checks if the key exists.
     ///
     /// # Examples
@@ -621,7 +599,7 @@ where
 
     /// Checks if the key exists.
     ///
-    /// It is an asynchronous method returning an `impl Future` for the caller to await or poll.
+    /// It is an asynchronous method returning an `impl Future` for the caller to await.
     ///
     /// # Examples
     ///
@@ -667,14 +645,16 @@ where
         // An acquire fence is required to correctly load the contents of the array.
         let mut current_array_ptr = self.array.load(Acquire, &barrier);
         while let Some(current_array_ref) = current_array_ptr.as_ref() {
-            if !current_array_ref.old_array(&barrier).is_null() {
-                current_array_ref.partial_rehash::<_, _, _, false>(
+            while !current_array_ref.old_array(&barrier).is_null() {
+                if current_array_ref.partial_rehash::<_, _, _>(
                     |key| self.hash(key),
                     |_, _| None,
+                    None,
                     &barrier,
-                );
-                current_array_ptr = self.array.load(Acquire, &barrier);
-                continue;
+                ) == Ok(true)
+                {
+                    break;
+                }
             }
 
             for cell_index in 0..current_array_ref.num_cells() {
@@ -708,51 +688,62 @@ where
     /// let hashmap: HashMap<usize, usize> = HashMap::default();
     ///
     /// let future_insert = hashmap.insert_async(1, 0);
-    /// let future_retain = hashmap.scan_async(|k, v| println!("{k} {v}"));
+    /// let future_scan = hashmap.scan_async(|k, v| println!("{k} {v}"));
     /// ```
     pub async fn scan_async<F: FnMut(&K, &V)>(&self, mut scanner: F) {
         // An acquire fence is required to correctly load the contents of the array.
-        let mut awaitable_barrier = AwaitableBarrier::default();
-        let mut current_array_holder = self.array.get_arc(Acquire, awaitable_barrier.barrier());
+        let mut current_array_holder = self.array.get_arc(Acquire, &Barrier::new());
         while let Some(current_array) = current_array_holder.take() {
-            while !current_array
-                .old_array(awaitable_barrier.barrier())
-                .is_null()
-            {
-                if current_array.partial_rehash::<_, _, _, true>(
+            while !current_array.old_array(&Barrier::new()).is_null() {
+                let mut async_wait = AsyncWait::default();
+                let mut async_wait_pinned = Pin::new(&mut async_wait);
+                if current_array.partial_rehash::<_, _, _>(
                     |key| self.hash(key),
                     |_, _| None,
-                    awaitable_barrier.barrier(),
-                ) {
-                    continue;
+                    Some(async_wait_pinned.mut_ptr()),
+                    &Barrier::new(),
+                ) == Ok(true)
+                {
+                    break;
                 }
-                awaitable_barrier.drop_barrier_and_yield().await;
+                async_wait_pinned.await;
             }
 
             for cell_index in 0..current_array.num_cells() {
-                loop {
+                let killed = loop {
+                    let mut async_wait = AsyncWait::default();
+                    let mut async_wait_pinned = Pin::new(&mut async_wait);
                     {
-                        // Limits the scope of `barrier`.
-                        let barrier = awaitable_barrier.barrier();
-                        if let Ok(result) =
-                            Locker::try_lock(current_array.cell(cell_index), barrier)
-                        {
+                        let barrier = Barrier::new();
+                        if let Ok(result) = Reader::try_lock_or_wait(
+                            current_array.cell(cell_index),
+                            async_wait_pinned.mut_ptr(),
+                            &barrier,
+                        ) {
                             if let Some(locker) = result {
-                                locker
-                                    .cell()
-                                    .iter(barrier)
-                                    .for_each(|((k, v), _)| scanner(k, v));
+                                let mut iterator = locker.cell().iter(&barrier);
+                                while iterator.next().is_some() {
+                                    if let Some((k, v)) = iterator.get() {
+                                        #[allow(clippy::cast_ref_to_mut)]
+                                        scanner(k, v);
+                                    }
+                                }
+                                break false;
                             }
-                            break;
-                        }
+
+                            // The `Cell` having been killed means that a new array has been
+                            // allocated.
+                            break true;
+                        };
                     }
-                    awaitable_barrier.drop_barrier_and_yield().await;
+                    async_wait_pinned.await;
+                };
+                if killed {
+                    break;
                 }
             }
 
-            if let Some(new_current_array) =
-                self.array.get_arc(Acquire, awaitable_barrier.barrier())
-            {
+            if let Some(new_current_array) = self.array.get_arc(Acquire, &Barrier::new()) {
                 if new_current_array.as_ptr() == current_array.as_ptr() {
                     break;
                 }
@@ -765,6 +756,8 @@ where
 
     /// Iterates over all the entries in the [`HashMap`].
     ///
+    /// This method allows modifying each value.
+    ///
     /// Key-value pairs that have existed since the invocation of the method are guaranteed to be
     /// visited if they are not removed, however the same key-value pair can be visited more than
     /// once if the [`HashMap`] gets resized by another thread.
@@ -795,11 +788,13 @@ where
 
     /// Iterates over all the entries in the [`HashMap`].
     ///
+    /// This method allows modifying each value.
+    ///
     /// Key-value pairs that have existed since the invocation of the method are guaranteed to be
     /// visited if they are not removed, however the same key-value pair can be visited more than
     /// once if the [`HashMap`] gets resized by another task.
     ///
-    /// It is an asynchronous method returning an `impl Future` for the caller to await or poll.
+    /// It is an asynchronous method returning an `impl Future` for the caller to await.
     ///
     /// # Examples
     ///
@@ -822,6 +817,8 @@ where
 
     /// Retains key-value pairs that satisfy the given predicate.
     ///
+    /// This method allows modifying each value.
+    ///
     /// Key-value pairs that have existed since the invocation of the method are guaranteed to be
     /// visited if they are not removed, however the same key-value pair can be visited more than
     /// once if the [`HashMap`] gets resized by another thread.
@@ -850,14 +847,16 @@ where
         // An acquire fence is required to correctly load the contents of the array.
         let mut current_array_ptr = self.array.load(Acquire, &barrier);
         while let Some(current_array_ref) = current_array_ptr.as_ref() {
-            if !current_array_ref.old_array(&barrier).is_null() {
-                current_array_ref.partial_rehash::<_, _, _, false>(
+            while !current_array_ref.old_array(&barrier).is_null() {
+                if current_array_ref.partial_rehash::<_, _, _>(
                     |key| self.hash(key),
                     |_, _| None,
+                    None,
                     &barrier,
-                );
-                current_array_ptr = self.array.load(Acquire, &barrier);
-                continue;
+                ) == Ok(true)
+                {
+                    break;
+                }
             }
 
             for cell_index in 0..current_array_ref.num_cells() {
@@ -897,12 +896,14 @@ where
 
     /// Retains key-value pairs that satisfy the given predicate.
     ///
+    /// This method allows modifying each value.
+    ///
     /// Key-value pairs that have existed since the invocation of the method are guaranteed to be
     /// visited if they are not removed, however the same key-value pair can be visited more than
     /// once if the [`HashMap`] gets resized by another task.
     ///
     /// It returns the number of entries remaining and removed. It is an asynchronous method
-    /// returning an `impl Future` for the caller to await or poll.
+    /// returning an `impl Future` for the caller to await.
     ///
     /// # Examples
     ///
@@ -922,33 +923,36 @@ where
         let mut num_removed: usize = 0;
 
         // An acquire fence is required to correctly load the contents of the array.
-        let mut awaitable_barrier = AwaitableBarrier::default();
-        let mut current_array_holder = self.array.get_arc(Acquire, awaitable_barrier.barrier());
+        let mut current_array_holder = self.array.get_arc(Acquire, &Barrier::new());
         while let Some(current_array) = current_array_holder.take() {
-            while !current_array
-                .old_array(awaitable_barrier.barrier())
-                .is_null()
-            {
-                if current_array.partial_rehash::<_, _, _, true>(
+            while !current_array.old_array(&Barrier::new()).is_null() {
+                let mut async_wait = AsyncWait::default();
+                let mut async_wait_pinned = Pin::new(&mut async_wait);
+                if current_array.partial_rehash::<_, _, _>(
                     |key| self.hash(key),
                     |_, _| None,
-                    awaitable_barrier.barrier(),
-                ) {
-                    continue;
+                    Some(async_wait_pinned.mut_ptr()),
+                    &Barrier::new(),
+                ) == Ok(true)
+                {
+                    break;
                 }
-                awaitable_barrier.drop_barrier_and_yield().await;
+                async_wait_pinned.await;
             }
 
             for cell_index in 0..current_array.num_cells() {
-                loop {
+                let killed = loop {
+                    let mut async_wait = AsyncWait::default();
+                    let mut async_wait_pinned = Pin::new(&mut async_wait);
                     {
-                        // Limits the scope of `barrier`.
-                        let barrier = awaitable_barrier.barrier();
-                        if let Ok(result) =
-                            Locker::try_lock(current_array.cell(cell_index), barrier)
-                        {
+                        let barrier = Barrier::new();
+                        if let Ok(result) = Locker::try_lock_or_wait(
+                            current_array.cell(cell_index),
+                            async_wait_pinned.mut_ptr(),
+                            &barrier,
+                        ) {
                             if let Some(locker) = result {
-                                let mut iterator = locker.cell().iter(barrier);
+                                let mut iterator = locker.cell().iter(&barrier);
                                 while iterator.next().is_some() {
                                     let retain = if let Some((k, v)) = iterator.get() {
                                         #[allow(clippy::cast_ref_to_mut)]
@@ -963,17 +967,22 @@ where
                                         num_removed = num_removed.saturating_add(1);
                                     }
                                 }
+                                break false;
                             }
-                            break;
-                        }
+
+                            // The `Cell` having been killed means that a new array has been
+                            // allocated.
+                            break true;
+                        };
                     }
-                    awaitable_barrier.drop_barrier_and_yield().await;
+                    async_wait_pinned.await;
+                };
+                if killed {
+                    break;
                 }
             }
 
-            if let Some(new_current_array) =
-                self.array.get_arc(Acquire, awaitable_barrier.barrier())
-            {
+            if let Some(new_current_array) = self.array.get_arc(Acquire, &Barrier::new()) {
                 if new_current_array.as_ptr() == current_array.as_ptr() {
                     break;
                 }
@@ -1010,7 +1019,7 @@ where
 
     /// Clears all the key-value pairs.
     ///
-    /// It is an asynchronous method returning an `impl Future` for the caller to await or poll.
+    /// It is an asynchronous method returning an `impl Future` for the caller to await.
     ///
     /// # Examples
     ///
