@@ -1,6 +1,5 @@
-//! The module implements [`HashSet`].
+//! [`HashSet`] is a concurrent and asynchronous hash set.
 
-use super::ebr::Barrier;
 use super::HashMap;
 
 use std::borrow::Borrow;
@@ -9,8 +8,7 @@ use std::hash::{BuildHasher, Hash};
 
 /// Scalable concurrent hash set.
 ///
-/// [`HashSet`] is a concurrent hash set based on [`HashMap`]. It internally uses a [`HashMap`]
-/// as its key container, thus sharing all the characteristics of [`HashMap`].
+/// [`HashSet`] is a concurrent and asynchronous hash set based on [`HashMap`].
 pub struct HashSet<K, H = RandomState>
 where
     K: 'static + Eq + Hash + Sync,
@@ -28,10 +26,6 @@ where
     ///
     /// The actual capacity is equal to or greater than the given capacity.
     ///
-    /// # Panics
-    ///
-    /// Panics if memory allocation fails.
-    ///
     /// # Examples
     ///
     /// ```
@@ -93,11 +87,6 @@ where
     ///
     /// Returns an error along with the supplied key if the key exists.
     ///
-    /// # Panics
-    ///
-    /// Panics if memory allocation fails, or the number of entries in the target cell reaches
-    /// `u32::MAX`.
-    ///
     /// # Examples
     ///
     /// ```
@@ -118,12 +107,13 @@ where
 
     /// Inserts a key into the [`HashSet`].
     ///
-    /// It is an asynchronous method returning an `impl Future` for the caller to await or poll.
+    /// It is an asynchronous method returning an `impl Future` for the caller to await.
     ///
     /// # Errors
     ///
     /// Returns an error along with the supplied key if the key exists.
     ///
+    /// function.
     /// # Examples
     ///
     /// ```
@@ -161,7 +151,7 @@ where
 
     /// Removes a key if the key exists.
     ///
-    /// It is an asynchronous method returning an `impl Future` for the caller to await or poll.
+    /// It is an asynchronous method returning an `impl Future` for the caller to await.
     ///
     /// # Examples
     ///
@@ -210,7 +200,7 @@ where
 
     /// Removes a key if the key exists and the given condition is met.
     ///
-    /// It is an asynchronous method returning an `impl Future` for the caller to await or poll.
+    /// It is an asynchronous method returning an `impl Future` for the caller to await.
     ///
     /// # Examples
     ///
@@ -253,19 +243,18 @@ where
     /// assert!(hashset.read(&1, |_| true).unwrap());
     /// ```
     #[inline]
-    pub fn read<Q, R, F: FnMut(&K) -> R>(&self, key_ref: &Q, reader: F) -> Option<R>
+    pub fn read<Q, R, F: FnMut(&K) -> R>(&self, key_ref: &Q, mut reader: F) -> Option<R>
     where
         K: Borrow<Q>,
         Q: Eq + Hash + ?Sized,
     {
-        let barrier = Barrier::new();
-        self.read_with(key_ref, reader, &barrier)
+        self.map.read(key_ref, |k, _| reader(k))
     }
 
     /// Reads a key.
     ///
     /// It returns `None` if the key does not exist. It is an asynchronous method returning an
-    /// `impl Future` for the caller to await or poll.
+    /// `impl Future` for the caller to await.
     ///
     /// # Examples
     ///
@@ -285,38 +274,6 @@ where
         self.map.read_async(key_ref, |k, _| reader(k)).await
     }
 
-    /// Reads a key using the supplied [`Barrier`].
-    ///
-    /// It returns `None` if the key does not exist.
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// use scc::ebr::Barrier;
-    /// use scc::HashSet;
-    ///
-    /// let hashset: HashSet<u64> = HashSet::default();
-    ///
-    /// assert!(hashset.insert(1).is_ok());
-    ///
-    /// let barrier = Barrier::new();
-    /// let key_ref = hashset.read_with(&1, |k| k, &barrier).unwrap();
-    /// assert_eq!(*key_ref, 1);
-    /// ```
-    #[inline]
-    pub fn read_with<'b, Q, R, F: FnMut(&'b K) -> R>(
-        &self,
-        key_ref: &Q,
-        mut reader: F,
-        barrier: &'b Barrier,
-    ) -> Option<R>
-    where
-        K: Borrow<Q>,
-        Q: Eq + Hash + ?Sized,
-    {
-        self.map.read_with(key_ref, |k, _| reader(k), barrier)
-    }
-
     /// Checks if the key exists.
     ///
     /// # Examples
@@ -341,7 +298,7 @@ where
 
     /// Checks if the key exists.
     ///
-    /// It is an asynchronous method returning an `impl Future` for the caller to await or poll.
+    /// It is an asynchronous method returning an `impl Future` for the caller to await.
     ///
     /// # Examples
     ///
@@ -400,7 +357,7 @@ where
     /// let hashset: HashSet<usize> = HashSet::default();
     ///
     /// let future_insert = hashset.insert_async(1);
-    /// let future_retain = hashset.scan_async(|k| println!("{k}"));
+    /// let future_scan = hashset.scan_async(|k| println!("{k}"));
     /// ```
     #[inline]
     pub async fn scan_async<F: FnMut(&K)>(&self, mut scanner: F) {
@@ -441,7 +398,7 @@ where
     /// they are not removed, however the same key can be visited more than once if the [`HashSet`]
     /// gets resized by another task.
     ///
-    /// It is an asynchronous method returning an `impl Future` for the caller to await or poll.
+    /// It is an asynchronous method returning an `impl Future` for the caller to await.
     ///
     /// # Examples
     ///
@@ -491,7 +448,7 @@ where
     /// gets resized by another task.
     ///
     /// It returns the number of entries remaining and removed. It is an asynchronous method
-    /// returning an `impl Future` for the caller to await or poll.
+    /// returning an `impl Future` for the caller to await.
     ///
     /// # Examples
     ///
@@ -527,7 +484,7 @@ where
 
     /// Clears all the keys.
     ///
-    /// It is an asynchronous method returning an `impl Future` for the caller to await or poll.
+    /// It is an asynchronous method returning an `impl Future` for the caller to await.
     ///
     /// # Examples
     ///
@@ -605,10 +562,6 @@ impl<K: 'static + Eq + Hash + Sync> Default for HashSet<K, RandomState> {
     ///
     /// The default hash builder is [`RandomState`], and the default capacity is `64`.
     ///
-    /// # Panics
-    ///
-    /// Panics if memory allocation fails.
-    ///
     /// # Examples
     ///
     /// ```
