@@ -94,11 +94,16 @@ mod hashmap_test {
                     assert!(result.is_none());
                 }
                 for id in range.clone() {
-                    if id % 5 == 0 {
+                    if id % 10 == 0 {
                         hashmap_cloned.upsert_async(id, || id, |_, v| *v = id).await;
-                    } else {
+                    } else if id % 5 == 0 {
+                        hashmap_cloned.upsert(id, || id, |_, v| *v = id);
+                    } else if id % 2 == 0 {
                         let result = hashmap_cloned.insert_async(id, id).await;
                         assert!(result.is_ok());
+                    } else {
+                        let result = hashmap_cloned.insert(id, id);
+                        assert!(result.is_ok());
                     }
                 }
                 for id in range.clone() {
@@ -119,10 +124,17 @@ mod hashmap_test {
                 for id in range.clone() {
                     let result = hashmap_cloned.read_async(&id, |_, v| *v).await;
                     assert_eq!(result, Some(id + 1));
+                    let result = hashmap_cloned.read(&id, |_, v| *v);
+                    assert_eq!(result, Some(id + 1));
                 }
                 for id in range.clone() {
-                    let result = hashmap_cloned.remove_if_async(&id, |v| *v == id + 1).await;
-                    assert_eq!(result, Some((id, id + 1)));
+                    if id % 2 == 0 {
+                        let result = hashmap_cloned.remove_if_async(&id, |v| *v == id + 1).await;
+                        assert_eq!(result, Some((id, id + 1)));
+                    } else {
+                        let result = hashmap_cloned.remove_if(&id, |v| *v == id + 1);
+                        assert_eq!(result, Some((id, id + 1)));
+                    }
                 }
                 for id in range {
                     let result = hashmap_cloned.remove_if_async(&id, |v| *v == id + 1).await;
@@ -980,3 +992,68 @@ mod treeindex_test {
         }
     }
 }
+
+#[cfg(test)]
+mod queue_test {
+    use crate::Queue;
+
+    use std::sync::atomic::AtomicUsize;
+    use std::sync::atomic::Ordering::Relaxed;
+    use std::sync::Arc;
+
+    use tokio::sync::Barrier as AsyncBarrier;
+
+    struct R(usize, usize);
+    impl R {
+        fn new(task_id: usize, seq: usize) -> R {
+            R(task_id, seq)
+        }
+    }
+
+    #[tokio::test(flavor = "multi_thread", worker_threads = 16)]
+    async fn mpmc() {
+        const NUM_TASKS: usize = 12;
+        const NUM_PRODUCERS: usize = NUM_TASKS / 2;
+        let workload_size = 256;
+        for _ in 0..256 {
+            let queue: Arc<Queue<R>> = Arc::new(Queue::default());
+            let num_popped: Arc<AtomicUsize> = Arc::new(AtomicUsize::default());
+            let mut task_handles = Vec::with_capacity(NUM_TASKS);
+            let barrier = Arc::new(AsyncBarrier::new(NUM_TASKS));
+            for task_id in 0..NUM_TASKS {
+                let barrier_cloned = barrier.clone();
+                let queue_cloned = queue.clone();
+                let num_popped_cloned = num_popped.clone();
+                task_handles.push(tokio::task::spawn(async move {
+                    barrier_cloned.wait().await;
+                    if task_id < NUM_PRODUCERS {
+                        for seq in 1..=workload_size {
+                            assert_eq!(queue_cloned.push(R::new(task_id, seq)).1, seq);
+                        }
+                    } else {
+                        let mut popped_acc: [usize; NUM_PRODUCERS] = Default::default();
+                        loop {
+                            let mut cnt = 0;
+                            while let Some(popped) = queue_cloned.pop() {
+                                cnt += 1;
+                                assert!(popped_acc[popped.0] < popped.1);
+                                popped_acc[popped.0] = popped.1;
+                            }
+                            if num_popped_cloned.fetch_add(cnt, Relaxed) + cnt
+                                == workload_size * NUM_PRODUCERS
+                            {
+                                break;
+                            }
+                            tokio::task::yield_now().await;
+                        }
+                    }
+                }));
+            }
+
+            for r in futures::future::join_all(task_handles).await {
+                assert!(r.is_ok());
+            }
+            assert!(queue.is_empty());
+        }
+    }
+}
