@@ -3,6 +3,7 @@ use super::leaf::{InsertResult, RemoveResult, Scanner};
 use super::leaf_node::{self, LeafNode};
 
 use crate::ebr::{Arc, AtomicArc, Barrier, Tag};
+use crate::wait_queue::AsyncWait;
 
 use std::borrow::Borrow;
 use std::sync::atomic::Ordering::{self, Acquire, Relaxed, Release};
@@ -32,26 +33,26 @@ where
     V: 'static + Clone + Send + Sync,
 {
     /// Creates a new [`InternalNode`].
-    pub fn new_internal_node() -> Node<K, V> {
+    pub(crate) fn new_internal_node() -> Node<K, V> {
         Node {
             node: Type::Internal(InternalNode::new()),
         }
     }
 
     /// Creates a new [`LeafNode`].
-    pub fn new_leaf_node() -> Node<K, V> {
+    pub(crate) fn new_leaf_node() -> Node<K, V> {
         Node {
             node: Type::Leaf(LeafNode::new()),
         }
     }
 
     /// Returns a reference to the node internal.
-    pub fn node(&self) -> &Type<K, V> {
+    pub(crate) fn node(&self) -> &Type<K, V> {
         &self.node
     }
 
     /// Returns the depth of the node.
-    pub fn depth(&self, depth: usize, barrier: &Barrier) -> usize {
+    pub(crate) fn depth(&self, depth: usize, barrier: &Barrier) -> usize {
         match &self.node {
             Type::Internal(internal_node) => internal_node.depth(depth, barrier),
             Type::Leaf(_) => depth,
@@ -59,7 +60,7 @@ where
     }
 
     /// Checks if the node has retired.
-    pub fn retired(&self, mo: Ordering) -> bool {
+    pub(crate) fn retired(&self, mo: Ordering) -> bool {
         match &self.node {
             Type::Internal(internal_node) => internal_node.retired(mo),
             Type::Leaf(leaf_node) => leaf_node.retired(mo),
@@ -67,7 +68,7 @@ where
     }
 
     /// Searches for an entry associated with the given key.
-    pub fn search<'b, Q>(&self, key: &Q, barrier: &'b Barrier) -> Option<&'b V>
+    pub(crate) fn search<'b, Q>(&self, key: &Q, barrier: &'b Barrier) -> Option<&'b V>
     where
         K: 'b + Borrow<Q>,
         Q: Ord + ?Sized,
@@ -80,8 +81,8 @@ where
 
     /// Returns the minimum key-value pair.
     ///
-    /// This method is not linearlizable.
-    pub fn min<'b>(&self, barrier: &'b Barrier) -> Option<Scanner<'b, K, V>> {
+    /// This method is not linearizable.
+    pub(crate) fn min<'b>(&self, barrier: &'b Barrier) -> Option<Scanner<'b, K, V>> {
         match &self.node {
             Type::Internal(internal_node) => internal_node.min(barrier),
             Type::Leaf(leaf_node) => leaf_node.min(barrier),
@@ -91,8 +92,12 @@ where
     /// Returns a [`Scanner`] pointing to an entry that is close enough to the entry with the
     /// maximum key among those keys smaller than or equal to the given key.
     ///
-    /// This method is not linearlizable.
-    pub fn max_le_appr<'b, Q>(&self, key: &Q, barrier: &'b Barrier) -> Option<Scanner<'b, K, V>>
+    /// This method is not linearizable.
+    pub(crate) fn max_le_appr<'b, Q>(
+        &self,
+        key: &Q,
+        barrier: &'b Barrier,
+    ) -> Option<Scanner<'b, K, V>>
     where
         K: 'b + Borrow<Q>,
         Q: Ord + ?Sized,
@@ -104,23 +109,25 @@ where
     }
 
     /// Inserts a key-value pair.
-    pub fn insert<const ASYNC: bool>(
+    pub(crate) fn insert(
         &self,
         key: K,
         value: V,
+        async_wait: Option<*mut AsyncWait>,
         barrier: &Barrier,
     ) -> Result<InsertResult<K, V>, (K, V)> {
         match &self.node {
-            Type::Internal(internal_node) => internal_node.insert::<ASYNC>(key, value, barrier),
-            Type::Leaf(leaf_node) => leaf_node.insert::<ASYNC>(key, value, barrier),
+            Type::Internal(internal_node) => internal_node.insert(key, value, async_wait, barrier),
+            Type::Leaf(leaf_node) => leaf_node.insert(key, value, async_wait, barrier),
         }
     }
 
     /// Removes an entry associated with the given key.
-    pub fn remove_if<Q, F: FnMut(&V) -> bool, const ASYNC: bool>(
+    pub(crate) fn remove_if<Q, F: FnMut(&V) -> bool>(
         &self,
         key: &Q,
         condition: &mut F,
+        async_wait: Option<*mut AsyncWait>,
         barrier: &Barrier,
     ) -> Result<RemoveResult, bool>
     where
@@ -129,14 +136,16 @@ where
     {
         match &self.node {
             Type::Internal(internal_node) => {
-                internal_node.remove_if::<_, _, ASYNC>(key, condition, barrier)
+                internal_node.remove_if::<_, _>(key, condition, async_wait, barrier)
+            }
+            Type::Leaf(leaf_node) => {
+                leaf_node.remove_if::<_, _>(key, condition, async_wait, barrier)
             }
-            Type::Leaf(leaf_node) => leaf_node.remove_if::<_, _, ASYNC>(key, condition, barrier),
         }
     }
 
     /// Splits the current root node.
-    pub fn split_root<const ASYNC: bool>(
+    pub(crate) fn split_root<const ASYNC: bool>(
         key: K,
         value: V,
         root: &AtomicArc<Node<K, V>>,
@@ -147,13 +156,14 @@ where
         let mut new_root: Node<K, V> = Node::new_internal_node();
         if let Type::Internal(internal_node) = &mut new_root.node {
             internal_node.unbounded_child = root.clone(Relaxed, barrier);
-            let result = internal_node.split_node::<ASYNC>(
+            let result = internal_node.split_node(
                 key,
                 value,
                 None,
                 root.load(Relaxed, barrier),
                 &internal_node.unbounded_child,
                 true,
+                None,
                 barrier,
             );
             let (key, value) = match result {
@@ -182,8 +192,9 @@ where
     /// # Errors
     ///
     /// Returns an error if a conflict is detected.
-    pub fn remove_root<const ASYNC: bool>(
+    pub(crate) fn remove_root(
         root: &AtomicArc<Node<K, V>>,
+        async_wait: Option<*mut AsyncWait>,
         barrier: &Barrier,
     ) -> Result<bool, ()> {
         let root_ptr = root.load(Acquire, barrier);
@@ -195,14 +206,14 @@ where
                     if let Some(locker) = internal_node::Locker::try_lock(internal_node, barrier) {
                         internal_node_locker.replace(locker);
                     } else {
-                        internal_node.wait::<ASYNC>(barrier);
+                        internal_node.wait(async_wait, barrier);
                     }
                 }
                 Type::Leaf(leaf_node) => {
                     if let Some(locker) = leaf_node::Locker::try_lock(leaf_node, barrier) {
                         leaf_node_locker.replace(locker);
                     } else {
-                        leaf_node.wait::<ASYNC>(barrier);
+                        leaf_node.wait(async_wait, barrier);
                     }
                 }
             };
@@ -232,7 +243,7 @@ where
     }
 
     /// Commits an on-going structural change.
-    pub fn commit(&self, barrier: &Barrier) {
+    pub(crate) fn commit(&self, barrier: &Barrier) {
         match &self.node {
             Type::Internal(internal_node) => internal_node.commit(barrier),
             Type::Leaf(leaf_node) => leaf_node.commit(barrier),
@@ -240,7 +251,7 @@ where
     }
 
     /// Rolls back an on-going structural change.
-    pub fn rollback(&self, barrier: &Barrier) {
+    pub(crate) fn rollback(&self, barrier: &Barrier) {
         match &self.node {
             Type::Internal(internal_node) => internal_node.rollback(barrier),
             Type::Leaf(leaf_node) => leaf_node.rollback(barrier),
@@ -250,7 +261,12 @@ where
     /// Cleans up logically deleted [`LeafNode`] instances in the linked list.
     ///
     /// If the target leaf node does not exist in the sub-tree, returns `false`.
-    pub fn cleanup_link<'b, Q>(&self, key: &Q, traverse_max: bool, barrier: &'b Barrier) -> bool
+    pub(crate) fn cleanup_link<'b, Q>(
+        &self,
+        key: &Q,
+        traverse_max: bool,
+        barrier: &'b Barrier,
+    ) -> bool
     where
         K: 'b + Borrow<Q>,
         Q: Ord + ?Sized,
